@@ -36,7 +36,7 @@
 #include <sys/time.h>
 #include <time.h>
 
-#include "xsi.h"
+#include <gpe/launch.h>
 
 #undef CHECK_RC
 
@@ -45,11 +45,13 @@ typedef struct _Key Key;
 struct _Key
 {
   int keycode;
+  int keycode2;
   unsigned int modifier;
   char *command;
   int type; /* 0 - on release; 1 - on late release; 2 - on press */
   Key *next;
   char *window;
+  int startup_notify;
 };
 
 struct key_event
@@ -206,6 +208,7 @@ create_new_key (char *key_string)
 
   k->modifier = 0;
   k->keycode = 0;
+  k->startup_notify = 1;
 
   key_str = strndup (key_string, sep1 - key_string);
 
@@ -225,6 +228,12 @@ create_new_key (char *key_string)
       k->command = strdup (&sep2[1]);
     }
 
+  if (k->command && k->command[0] == '~')
+    {
+      k->command++;
+      k->startup_notify = 0;
+    }
+
   if (strlen (key_str) < 4)
     {
       fprintf (stderr, "No key specified: '%s'\n", key_string);
@@ -260,9 +269,23 @@ create_new_key (char *key_string)
       k->keycode = XKeysymToKeycode (dpy, XStringToKeysym (key_str + 8));
       k->type = 1;
     }
+  else if ((strlen (key_str) > 11) && (strncmp (key_str+3, "Combine ", 8) == 0))
+    {
+      char *sep = strchr (key_str + 11, ' ');
+      if (sep)
+	{
+	  *sep = 0;
+	  k->keycode2 = XKeysymToKeycode (dpy, XStringToKeysym (sep + 1));
+	}
+      else
+	k->keycode2 = 0;
+      k->keycode = XKeysymToKeycode (dpy, XStringToKeysym (key_str + 11));
+      k->type = 4;
+    }
   else
     {
-      k->keycode = XKeysymToKeycode (dpy, XStringToKeysym (key_str + 3));
+      if (key_str && strlen (key_str))
+        k->keycode = XKeysymToKeycode (dpy, XStringToKeysym (g_strstrip(key_str + 3)));
       k->type = 0;
     }
 
@@ -287,6 +310,8 @@ create_new_key (char *key_string)
     }
 
   grab_key (k->keycode, k->modifier, root);
+  if (k->type == 4 && k->keycode2)
+    grab_key (k->keycode2, k->modifier, root);
 
   free (key_str);
 
@@ -508,6 +533,20 @@ static void print_key (XEvent ev)
 }
 #endif
 
+int try_to_raise_window (Display *display, char *binary)
+{
+  Window w;
+
+  w = gpe_launch_get_window_for_binary (display, binary);
+  if (w != None)
+    {
+      gpe_launch_activate_window (display, w);
+      return 0;
+    }
+
+  return -1;
+}
+
 /*
  *
  * Main
@@ -538,24 +577,82 @@ void process_key (XEvent ev, int type)
 		    printf ("*** faking key event following pressed event ***\n");
 #endif
 		  }
-		else if (k->window == NULL
-		    || (try_to_raise_window (dpy, k->window) != 0))
+		else if (try_to_raise_window (dpy, k->command) != 0
+		  && !gpe_launch_startup_is_pending (dpy, k->command))
 		  {
 #ifdef DEBUG
 		    printf ("running program\n");
 #endif
-		    fork_exec (k->command);
+		    gpe_launch_program_with_callback (dpy, k->command, k->command, TRUE, NULL, NULL);
 		  }
 
 		k = lastkey;
 	      }
 }
 
+void
+process_combine (void)
+{
+  Key *k;
+
+  for (k = key; k != NULL; k = k->next)
+    {
+      if (k->type == 4)
+	{
+	  struct key_event **kp;
+	  struct key_event **ka = NULL, **kb = NULL;
+
+	  for (kp = &keys_down; *kp; kp = &(*kp)->next)
+	    {
+	      struct key_event *ke = *kp;
+	      if (ke->ev.xkey.keycode == k->keycode)
+		ka = kp;
+	      else if (ke->ev.xkey.keycode == k->keycode2)
+		kb = kp;
+	    }
+
+	  if (ka && kb)
+	    {
+	      XEvent ev;
+	      struct key_event *kpa, *kpb;
+
+	      kpa = *ka;
+	      kpb = *kb;
+
+	      if (kpa->next == kpb)
+		{
+		  *ka = kpb->next;
+		}
+	      else if (kpb->next == kpa)
+		{
+		  *kb = kpa->next;
+		}
+	      else
+		{
+		  *ka = kpa->next;
+		  *kb = kpb->next;
+		}
+
+	      free (kpa);
+	      free (kpb);
+
+	      ev.xkey.state = 0;
+	      ev.xkey.keycode = k->keycode;
+	      
+	      process_key (ev, 4);
+	    }
+	}
+    }
+}
+
 int
 main (int argc, char *argv[])
 {
   initialize (argc, argv);
 
+  XSelectInput (dpy, DefaultRootWindow (dpy), PropertyChangeMask);
+  gpe_launch_monitor_display (dpy);
+
   signal (SIGCHLD, SIG_IGN);
 
   while (1)
@@ -591,6 +688,10 @@ main (int argc, char *argv[])
 	}
 
       XNextEvent (dpy, &ev);
+
+      if (gpe_launch_process_event (&ev))
+	continue;
+      
       if (ev.type == KeyPress)
         {
 	  struct key_event *k;
@@ -603,36 +704,19 @@ main (int argc, char *argv[])
 
 	  gettimeofday (&time_now, NULL);
 
-	  timersub (&time_now, &key_press_time, &time_elapsed);
-	  if ((time_elapsed.tv_sec == 0) && (time_elapsed.tv_usec < 500000))
-	    {
-#ifdef DEBUG
-	      fprintf (stderr, "ignoring - too soon after previous\n");
-#endif
-	      key_press_time.tv_sec = 0;
-	      key_press_time.tv_usec = 0;
-	    }
-          else
-	    {
-	      key_press_time = time_now;
-	      process_key (ev, 2);
-	    }
+	  key_press_time = time_now;
+	  process_key (ev, 2);
 
 	  k = malloc (sizeof (struct key_event));
 	  memcpy (&k->ev, &ev, sizeof (ev));
-	  k->next = NULL;
 	  time_elapsed.tv_sec = 0;
 	  time_elapsed.tv_usec = 500000;
 	  timeradd (&time_now, &time_elapsed, &k->time);
 
-	  if (keys_down)
-	    {
-	      struct key_event *p;
-	      for (p = keys_down; p->next; p = p->next)
-		p->next = k;
-	    }
-	  else
-	    keys_down = k;
+	  k->next = keys_down;
+	  keys_down = k;
+
+	  process_combine ();
 	}
       else if (ev.type == KeyRelease
 	       && (key_press_time.tv_sec || key_press_time.tv_usec))
